#include "FullWindowRenderTest.h"
#include <windows.h>
#include <process.h>
#include <commctrl.h>
#include <unknwn.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <tchar.h>
#include <Uxtheme.h>
#include <Dwmapi.h>
#include <windowsx.h>

#include "DeviceManagerD3D9.h"

#define MAX_LOADSTRING 100

HWND mWnd = NULL;
HWND mFocusWnd = NULL;
HINSTANCE hInst;
TCHAR szTitle[MAX_LOADSTRING];
TCHAR szWindowClass[MAX_LOADSTRING];

ATOM				MyRegisterClass(HINSTANCE hInstance);
BOOL				InitInstance(HINSTANCE, int);
LRESULT CALLBACK	WndProc(HWND, UINT, WPARAM, LPARAM);

int APIENTRY _tWinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPTSTR    lpCmdLine,
                     int       nCmdShow)
{
	UNREFERENCED_PARAMETER(hPrevInstance);
	UNREFERENCED_PARAMETER(lpCmdLine);

	MSG msg;

	LoadString(hInstance, IDS_APP_TITLE, szTitle, MAX_LOADSTRING);
	LoadString(hInstance, IDC_FULLWINDOWRENDERTEST, szWindowClass, MAX_LOADSTRING);
	MyRegisterClass(hInstance);

	if (!InitInstance (hInstance, nCmdShow))
		return FALSE;

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return (int) msg.wParam;
}

ATOM MyRegisterClass(HINSTANCE hInstance)
{
	WNDCLASSEX wcex;

	wcex.cbSize = sizeof(WNDCLASSEX);

	wcex.style			= CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc	= WndProc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= LoadIcon(hInstance, MAKEINTRESOURCE(IDI_FULLWINDOWRENDERTEST));
	wcex.hCursor		= LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground	= (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName	= MAKEINTRESOURCE(IDC_FULLWINDOWRENDERTEST);
	wcex.lpszClassName	= szWindowClass;
	wcex.hIconSm		= LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

	return RegisterClassEx(&wcex);
}

static DeviceManagerD3D9 *mDeviceManager = NULL;
static SwapChainD3D9* mSwapChain = NULL;
static DeviceManagerD3D9* deviceManager() { return mDeviceManager; }

BOOL InitInstance(HINSTANCE hInstance, int nCmdShow)
{
  hInst = hInstance; // Store instance handle in our global variable

  mWnd = CreateWindow(szWindowClass, szTitle, WS_OVERLAPPEDWINDOW,
  CW_USEDEFAULT, 0, 300, 300, NULL, NULL, hInstance, NULL);

  if (!mWnd)
  {
    return FALSE;
  }

  mDeviceManager = new DeviceManagerD3D9();
  mDeviceManager->Init();

  if (!mDeviceManager->Init()) {
    mDeviceManager = nsnull;
    return PR_FALSE;
  }

  mSwapChain = mDeviceManager->CreateSwapChain(mWnd);

  ShowWindow(mWnd, nCmdShow);
  UpdateWindow(mWnd);

  return TRUE;
}

void SetupPipeline()
{
  /*  
  RECT r;
  GetClientRect(mWnd, &r);
  
  gfx3DMatrix viewMatrix;
  viewMatrix._11 = 2.0f / rect.width;
  viewMatrix._22 = -2.0f / rect.height;
  viewMatrix._41 = -1.0f;
  viewMatrix._42 = 1.0f;

  
  HRESULT hr = device()->SetVertexShaderConstantF(CBmProjection,
                                                  &viewMatrix._11, 4);

  if (FAILED(hr)) {
    NS_WARNING("Failed to set projection shader constant!");
  }

  hr = device()->SetVertexShaderConstantF(CBvTextureCoords,
                                          ShaderConstantRect(0, 0, 1.0f, 1.0f),
                                          1);

  if (FAILED(hr)) {
    NS_WARNING("Failed to set texCoords shader constant!");
  }

  float offset[] = { 0, 0, 0, 0 };
  hr = device()->SetVertexShaderConstantF(CBvRenderTargetOffset, offset, 1);

  if (FAILED(hr)) {
    NS_WARNING("Failed to set RenderTargetOffset shader constant!");
  }
  */
}

MARGINS mMargins = { 30, 30, 32, 30 }; // left right top bottom

void Render()
{
  if (!mSwapChain->PrepareForRendering()) {
    return;
  }

  deviceManager()->SetupRenderState();

  SetupPipeline();

  RECT r;
  GetClientRect(mWnd, &r);
  r.left += mMargins.cxLeftWidth;
  r.top += mMargins.cyTopHeight;
  r.bottom += 2;

  deviceManager()->device()->BeginScene();

  deviceManager()->device()->SetScissorRect(&r);

  //static_cast<LayerD3D9*>(mRoot->ImplData())->RenderLayer();

  deviceManager()->device()->EndScene();

  deviceManager()->device()->Clear(0, NULL, D3DCLEAR_TARGET, 0xFFA000A0, 0, 0);

  mSwapChain->Present(r);
}

void UpdateOpaqueRegion()
{
  DwmExtendFrameIntoClientArea(mWnd, &mMargins);
  DWMNCRENDERINGPOLICY policy = DWMNCRP_USEWINDOWSTYLE;
  DwmSetWindowAttribute(mWnd, DWMWA_NCRENDERING_POLICY, &policy, sizeof policy);
}

DWORD
ClientMarginHitTestPoint(int mx, int my)
{
  // Calculations are done in screen coords
  RECT winRect;
  GetWindowRect(mWnd, &winRect);

  DWORD testResult = HTCLIENT;

  bool top    = false;
  bool bottom = false;
  bool left   = false;
  bool right  = false;

  int sideMargin = 8;
  int titlebarMargin = 32;

  if (my >= winRect.top && my < (winRect.top + titlebarMargin))
    top = PR_TRUE;
  else if (my < winRect.bottom && my >= (winRect.bottom - sideMargin))
    bottom = PR_TRUE;

  if (mx >= winRect.left && mx < (winRect.left +
                                  (bottom ? (2*sideMargin) : sideMargin)))
    left = PR_TRUE;
  else if (mx < winRect.right && mx >= (winRect.right -
                                        (bottom ? (2*sideMargin) : sideMargin)))
    right = PR_TRUE;

  if (top) {
    testResult = HTTOP;
    if (left)
      testResult = HTTOPLEFT;
    else if (right)
      testResult = HTTOPRIGHT;
  } else if (bottom) {
    testResult = HTBOTTOM;
    if (left)
      testResult = HTBOTTOMLEFT;
    else if (right)
      testResult = HTBOTTOMRIGHT;
  } else {
    if (left)
      testResult = HTLEFT;
    if (right)
      testResult = HTRIGHT;
  }
  return testResult;
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	PAINTSTRUCT ps;
	HDC hdc;

  LRESULT dwmHitResult;
  if (DwmDefWindowProc(hWnd, message, wParam, lParam, &dwmHitResult))
    return dwmHitResult;
  
	switch (message)
	{
    case WM_NCHITTEST:
      return ClientMarginHitTestPoint(GET_X_LPARAM(lParam), GET_Y_LPARAM(lParam));
    
    case WM_NCCALCSIZE:
    {
      if (!wParam)
        return 0;

      // before:
      // rgrc[0]: the proposed new window coordinates
      // rgrc[1]: the current window cordinates
      // rgrc[2]: the source client area coordinates
      // pncsp->lppos: move/size data
      // after:
      // rgrc[0]: the new client coordinates
      // rgrc[1]: the client destination coordinates
      // rgrc[2]: the client source coordinates
      NCCALCSIZE_PARAMS *pncsp = reinterpret_cast<NCCALCSIZE_PARAMS*>(lParam);
      
      return 0;
  }
  
	case WM_PAINT:
		hdc = BeginPaint(hWnd, &ps);
		Render();
		UpdateOpaqueRegion();

    /*
    RECT r;
    GetClientRect(mWnd, &r);
    r.left += mMargins.cxLeftWidth;
    r.top += mMargins.cyTopHeight;
    r.bottom = r.top + 20;
    r.right = r.left + 20;
    FillRect(hdc, &r, (HBRUSH) (COLOR_HIGHLIGHTTEXT+1));
    */
		EndPaint(hWnd, &ps);
		break;

  case WM_ERASEBKGND:
  {
    RECT rect;
    GetClientRect(mWnd, &rect);
    FillRect((HDC)wParam, &rect, (HBRUSH)GetStockObject (BLACK_BRUSH));
    break;
  }

	case WM_DESTROY:
		PostQuitMessage(0);
		break;
	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

